from GOESLayer import GOESLayer
import logging

##
# @brief CADU Layer class.
# @details First layer of LRIT packet processing.  Performs derandomization
# and inversion, if necessary.
# @author Miguel Dominguez
# @date 2016
class CADULayer(GOESLayer): 
   CADU_LENGTH_BYTES = 1020
   HEADER_LENGTH = 4
   BITS_PER_BYTE = 8
   MAGIC_WORD = 0x1ACFFC1D
   MAGIC_WORD_INV = (~MAGIC_WORD) & 0XFFFFFFFF
   ##
   # @fn __init__
   # @brief Constructor.
   # @details Creates a logger and initializes derandomization table.
   def __init__(self):
      GOESLayer.__init__(self)
      self.logger = logging.getLogger(__name__)
      self.initPnSequence()

   ##
   # @fn initPnSequence
   # @brief Generates derandomization table.
   # @details Derandomization is done by using an 8-bit linear feedback
   # shift register, intialized to 255, and with a polynomial of
   # x^8 + x^7 + x^5 + x^3 + 1
   def initPnSequence(self):
      register = 0xff #Initial shift register is all ones
      outputBit = 0
      #Generate array of bytes used to xor with data to derandomize
      self.pnSequence = [0 for i in range(self.CADU_LENGTH_BYTES)] 
      for i in range(0,self.CADU_LENGTH_BYTES):
         for j in range(0,self.BITS_PER_BYTE):
            self.pnSequence[i] <<= 1
            self.pnSequence[i] |= ((register & 0x80) >> 7)
            feedbackBit = (((register & 0x80)^(0x80 & (register << 3)))^(0x80 & (register << 5)))^(0x80 & (register << 7))
            feedbackBit = feedbackBit >> 7
            register = ((register << 1) &0xff) | feedbackBit
      
   ##
   # @fn processData
   # @brief Processes input frames as raw bytes, outputting derandomized data.
   # @detail First a preamble is checked to see if it contains the required magic
   # word.  If it contains either the exact vaue or its exact bit inverse, the data
   # is derandomized and passed forward, without the preamble.  If the preamble is inverted,
   # all of the bytes in the frame are inverted before derandomization.
   # @param[in] data 1024 unsigned bytes.
   # @param[in] opsMsgs Some sort of filler data, not used
   # @param[out] outData 1020 derandomized unsigned bytes (preamble stripped away)
   # @param[out] outOpsMsg Filler, not used
   def processData(self,data,opsMsg):
      relevantData = data[self.HEADER_LENGTH:len(data)]
      #Throw away 4-byte header
      magicWordCheck = 0
      magicWordFound = False
      for i in range(self.HEADER_LENGTH):
         magicWordCheck <<= self.BITS_PER_BYTE
         magicWordCheck |= data[i]

      if magicWordCheck == self.MAGIC_WORD_INV:
         self.logger.info('Inverse magic word found! {:08x}'.format(magicWordCheck))
         magicWordFound = True
         relevantData = [(~x) & 0xFF for x in relevantData]
      elif magicWordCheck == self.MAGIC_WORD:
         magicWordFound = True
         self.logger.info('Normal magic word found! {:08x}'.format(magicWordCheck))

      if magicWordFound:
         relevantData = [relevantData[i]^self.pnSequence[i] for i in range(0,self.CADU_LENGTH_BYTES)]
         return (relevantData,True)
      else:
         self.logger.info('No Magic Word: {:08x}: '.format(magicWordCheck))
         return([],False)
