from GOESLayer import GOESLayer
import sys
import os
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../reedsolomon/unireedsolomon'))
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../reedsolomon'))
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/frameTypes'))
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/opsMsgs'))
import rs,ff
from BerlekampTransform import Berlekamp
from itertools import product
from VCDUMsg import VCDUMsg
from VCDUOpsMsg import VCDUOpsMsg
import logging
import multiprocessing

##
# @brief VCDU (Virtual Channel) Layer class.
# @details Second layer of LRIT packet processing. Deinterleaves and performs
# Reed-Solomon Decoding.
# @warning The RS decoding is too slow for HRIT speeds!  Based on a 3rd-party
# library called 'unireedsolomon.' Should probably be implemented in C wrapped
# in Python.  Fortuitously GNURadio has CCSDS RS decoder implementations in C!
# @author Miguel Dominguez
# @date 2016
class VCDULayer(GOESLayer): 

   ##
   # @fn __init__
   # @brief Constructor.
   # @details Creates a logger and initializes Reed-Solomon Decoder
   def __init__(self):
      GOESLayer.__init__(self)
      self.logger = logging.getLogger(__name__)
      self.INTERLEAVE_CHANNELS = 4
      self.initRSCoder()

   ##
   # @fn initRSCoder
   # @brief Initialize the RS decoder with CCSDS parameters
   # @details Developer note: The lookup table is used because the
   # CCSDS polynomial root is 2^11 (modulo some number), which is definitely
   # unusual
   def initRSCoder(self):
      self.ALPHA = 2
      self.G = 11
      self.F = 0x187
      self.M = 8
      self.N = 255
      self.K = 223
      self.E = (self.N - self.K)/2
      self.J0 = 2**(self.M - 1)  - self.E

      [GF2int_exptable, GF2int_logtable] = ff.init_lut(generator=self.ALPHA,prim=self.F,c_exp=self.M)

      self.rsCoder = rs.RSCoder(self.N,self.K,generator=ff.GF2int(self.ALPHA)**self.G,prim=self.F,fcr=self.J0)

   ##
   # @fn deInterleave
   # @brief Deinterleave the VCDU Data
   # @details This looks complicated because there are 4 interleaved channels,
   # plus the data bits and parity bits are separated in the frame. In the future
   # this should be done in numpy
   # @param[in] data Unsigned bytes representing the RS-coded VCDU frame
   # @param[out] outData An unsigned byte array of the 4 RS codewords, deinterleaved
   def deInterleave(self,data):
      dataDeInterleaved = [data[i + n] for (n,i) in product(range(self.INTERLEAVE_CHANNELS),range(0,len(data),self.INTERLEAVE_CHANNELS))]
      dataBytesSplit = [dataDeInterleaved[i:(i + self.N)] for i in range(0,len(dataDeInterleaved),self.N)]
      return dataBytesSplit 

   ##
   # @fn processData
   # @brief Deinterleave and RS-decode data
   # @detail If data has unfixable RS errors, the data is not passed forward
   # and a note is made in the Ops Message to that affect.  Also, Berlekamp
   # Basis transforms are used to properly format data for RS decode
   # @param[in] data 1020 unsigned bytes.
   # @param[in] opsMsgs Some sort of filler data, not used
   # @param[out] outData VCDUMsg Object
   # @param[out] outOpsMsg VCDUOpsMsg object, mentions failed RS decodings and
   # whether the data is marked as fill.
   def processData(self,data,opsMsg):
      output = VCDUMsg()
      opsMsg = VCDUOpsMsg()
      codewords = self.deInterleave(data)
      errorCorrectedData = []
      mergedData = []
      for i in range(self.INTERLEAVE_CHANNELS):
         try:
            msg,ecc = self.rsCoder.decode_fast(Berlekamp.InverseTransform(codewords[i]),return_string=False,nostrip=True)
            errorCorrectedData.append([(int(x) & 0xFF) for x in Berlekamp.Transform(msg)])
         except rs.RSCodecError:
            #TODO: LOG THE PROBLEM
            opsMsg.RSFail(i)

      #Only package up the frame if it passes Reed Solomon decoding
      if opsMsg.RSNoErrors():
         mergedData = [errorCorrectedData[n][i] for (i,n) in product(range(self.K),range(self.INTERLEAVE_CHANNELS))]
         output.fromFrame(mergedData)
#         self.logger.debug(output)
#         print("VCDU Header Count: {0}".format(output.counter))
         #Get rid of fill VCDUs, no interesting data contained therein
         if output.vcID == VCDUMsg.FILL_VCID:
            opsMsg.setIsFill(True)
      else:
         output = []
         failedChannels = [str(chan) for chan in opsMsg.RSErrorLocations()]
         self.logger.warn('VCDU frame contained unfixable Reed Solomon Errors in channels: {0}'.format(", ".join(failedChannels)))
      return (output,opsMsg)
