from GOESFrame import GOESFrame

##
# @brief VCDU Message class.
# @details Contains a single Virtual Channel Data Unit. Most interesting
# out of the header is the VCID. The protocol features multiple Virtual
# Channels through which different files can be received in parallel.
# The "data" is essentially the raw unsigned bytes for the MPDU
# @author Miguel Dominguez
# @date 2016
class VCDUMsg(GOESFrame):
    VERSION_NUM_MASK = 0xC0
    VERSION_NUM_SHIFT = 6
    SCID_MASK_LEFT = 0x3F
    SCID_LEFT_SHIFT = 2
    SCID_MASK_RIGHT = 0xC0
    SCID_RIGHT_SHIFT = 6
    VCID_MASK = 0x3F
    REPLAY_FLAG_MASK = 0x80
    REPLAY_FLAG_SHIFT = 7
    SPARE_MASK = 0x7F

    HEADER_SIZE_BYTES = 6
    DATA_SIZE_BYTES = 886
    INTERLEAVE_COUNT = 4
    CHAN_SIZE_BYTES = DATA_SIZE_BYTES/INTERLEAVE_COUNT
    MAX_VCDU_COUNTER = 2 ** 24

    FILL_VCID = 63

    ##
    # @fn getHeader
    # @brief Extracts the header values with appropriate bit masking and shifting
    # @param[in] headerBytes Unsigned byte list sufficiently long enough to
    # contain the header (6 bytes)
    def getHeader(self, headerBytes):
        self.versionNum = (headerBytes[0] & self.VERSION_NUM_MASK) >> self.VERSION_NUM_SHIFT
        self.scID = ((headerBytes[0] & self.SCID_MASK_LEFT) << self.SCID_LEFT_SHIFT) | ((headerBytes[1] & self.SCID_MASK_RIGHT) >> self.SCID_RIGHT_SHIFT)
        self.vcID = (headerBytes[1] & self.VCID_MASK)
        self.counter = (headerBytes[2] << 16) | (headerBytes[3] << 8) | (headerBytes[4])
        self.replayFlag = (headerBytes[5] & self.REPLAY_FLAG_MASK) >> self.REPLAY_FLAG_SHIFT
        self.spare = (headerBytes[5] & self.SPARE_MASK)

    ##
    # @fn fromFrame
    # @brief Extracts the VCDU Message from raw unsigned bytes
    # @details This also calls getHeader
    # @param[in] data 1020 raw unsigned bytes
    def fromFrame(self,data):
        self.getHeader(data[0:self.HEADER_SIZE_BYTES])
        self.data = data[self.HEADER_SIZE_BYTES:len(data)]

    ##
    # @fn calcBER
    # @brief Calculates BER from fill packets
    def calcBER(self):
        onesum = 0
        for n in self.data:
            onesum += bin(n).count("1")
        self.ber = 1.0*onesum/len(self.data)


    ##
    # @fn __str__
    # @brief Returns the Header information as a string
    # @param[out] str Header information as a string
    def __str__(self):
        output = "Version Number: {0}\n".format(self.versionNum)
        output += "S/C ID: {0}\n".format(self.scID)
        output += "VC ID: {0}\n".format(self.vcID)
        output += "Counter: {0}\n".format(self.counter)
        output += "Replay Flag: {0}".format(self.replayFlag)
        #output += "Data:\n"
        #output += "".join([chr(x) for x in self.data])

        return output
